import numbers
import numpy as np
import scipy.constants
from scipy.linalg import block_diag
from scipy.special import sph_harm
from sympy.physics.wigner import wigner_3j

from diatomic import log_time

"""
Basis function labels and order
"""


class HalfInt:
    """
    Container to encapsulate half-integer spin & projection lengths. Values will be
    automatically downcast to an integer if the numerator of <of>/2 is an even number
    as soon as possible, even upon initialisation.
    """

    def __new__(cls, *, of):
        if not isinstance(of, numbers.Integral):
            raise TypeError("The argument 'of' must be an integer.")
        elif of % 2 == 0:
            return of // 2
        else:
            return super().__new__(cls)

    def __init__(self, *, of):
        self._double = of

    def _fraction_str(self):
        return f"({self._double}/2)"

    def __repr__(self):
        return f"({self._fraction_str()} : {self.__class__.__name__})"

    def __str__(self):
        return self._fraction_str()

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self._double == other._double
        elif isinstance(other, int):
            return self._double == 2 * other
        elif isinstance(other, float):
            return self.__float__() == other
        else:
            return NotImplemented

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        if isinstance(other, self.__class__):
            return self._double < other._double
        elif isinstance(other, int):
            return self._double < 2 * other
        elif isinstance(other, float):
            return self.__float__() < other
        else:
            return NotImplemented

    def __le__(self, other):
        return self < other or self == other

    def __gt__(self, other):
        if isinstance(other, self.__class__):
            return self._double > other._double
        elif isinstance(other, int):
            return self._double > 2 * other
        elif isinstance(other, float):
            return self.__float__() > other
        else:
            return NotImplemented

    def __ge__(self, other):
        return self > other or self == other

    def __float__(self):
        return self._double / 2

    def __int__(self):
        return int(self._double / 2)

    def __abs__(self):
        return self.__class__(of=abs(self._double))

    def __add__(self, other):
        if isinstance(other, self.__class__):
            return self.__class__(of=self._double + other._double)
        elif isinstance(other, int):
            return self.__class__(of=self._double + 2 * other)
        elif isinstance(other, float):
            return self.__float__() + other
        else:
            return NotImplemented

    def __radd__(self, other):
        return self + other

    def __sub__(self, other):
        if isinstance(other, self.__class__):
            return self.__class__(of=self._double - other._double)
        elif isinstance(other, int):
            return self.__class__(of=self._double - 2 * other)
        elif isinstance(other, float):
            return self.__float__() - other
        else:
            return NotImplemented

    def __rsub__(self, other):
        return -(self - other)

    def __neg__(self):
        return self.__class__(of=-self._double)

    def __mul__(self, other):
        if isinstance(other, self.__class__):
            return self.__float__() * float(other)
        elif isinstance(other, int):
            return self.__class__(of=self._double * other)
        elif isinstance(other, float):
            return self.__float__() * other
        else:
            return NotImplemented

    def __rmul__(self, other):
        return self * other


def proj_iter(angmom):
    """
    Generator function that yields the projections of angular momentum quantum numbers.

    This function takes an angular momentum value (angmom) and generates a sequence of
    projection quantum numbers (m) from +angmom to -angmom in steps of 1 angular
    momentum unit, represented as HalfInt objects.

    Args:
        angmom (int | HalfInt | float): The angular momentum quantum number,
            which can be an integer, HalfInt, or float.

    Yields:
        HalfInt: A projection quantum number of the angular momentum starting from
            +angmom to -angmom in decrements of 1 (in terms of angular momentum units).
    """
    return (HalfInt(of=dm) for dm in range(int(2 * angmom), -(int(2 * angmom) + 1), -2))
    # return (HalfInt(double=dm) for dm in range(-2 * angmom, 2 * angmom + 1, 2))


def sph_iter(Nmax):
    """
    Generator function that yields tuples of spherical harmonic angular momentum
    quantum numbers up to a specified maximum angular momentum Nmax.

    Args:
        Nmax (int): The maximum (inclusive) angular momentum quantum number to
            iterate up to.

    Yields:
        tuple: A tuple of two elements where the first element is the angular momentum
            quantum number (N) and the second element is the projection quantum number
            (MN) as yielded by the proj_iter function for that N.
    """
    return ((N, MN) for N in range(0, Nmax + 1) for MN in proj_iter(N))


def uncoupled_basis_iter(Nmax, I1, I2):
    """
    Generator function that yields a Cartesian product of tuples representing the
    uncoupled basis states  for two angular momenta up to a specified maximum angular
    momentum.

    Args:
        Nmax (int): The maximum angular momentum quantum number for the orbital angular
            momentum.
        I1 (int | HalfInt | float): The angular momentum quantum number for the first
            intrinsic angular momentum of the nucleus.
        I2 (int | HalfInt | float): The angular momentum quantum number for the second
            intrinsic angular momentumof the nucleus.

    Yields:
        tuple: A tuple of four elements (N, MN, M1, M2), where N is an angular momentum
            quantum number from 0 to Nmax (inclusive), MN is the projection of N,
            and M1 and M2 are the projections of I1 and I2, respectively.
    """
    return (
        (N, MN, M1, M2)
        for (N, MN) in sph_iter(Nmax)
        for M1 in proj_iter(I1)
        for M2 in proj_iter(I2)
    )


def uncoupled_basis_pos(N, MN, MI1, MI2, I1, I2):
    """
    Calculates the position (index) of a given uncoupled basis state in a linear array
    that would be produced by uncoupled_basis_iter.

    Args:
        N (int | HalfInt | float): The angular momentum quantum number for the first
            component of the basis state.
        MN (int | HalfInt | float): The projection quantum number of N.
        MI1 (int | HalfInt | float): The projection quantum number of the first
            intrinsic angular momentum I1.
        MI2 (int | HalfInt | float): The projection quantum number of the second
            intrinsic angular momentum I2.
        I1 (int | HalfInt | float): The angular momentum quantum number for the
            first intrinsic angular momentum.
        I2 (int | HalfInt | float): The angular momentum quantum number for the
            second intrinsic angular momentum.

    Returns:
        int: The index position of the uncoupled basis state in a linear array.
    """
    return int(
        (I2 - MI2) + (2 * I2 + 1) * ((I1 - MI1) + (2 * I1 + 1) * (N**2 + N - MN))
    )


def num_proj_with_below(Nmax):
    """
    Calculates the total number of projection quantum numbers for all angular momentum
    quantum numbers from 0 up to a specified maximum value, Nmax.

    The function sums analyticalls the number of possible projections for each
    angular momentum quantum number, which follows the formula 2 * N + 1.
    i.e. `sum([2 * x + 1 for x in range(0, Nmax + 1)])`

    Args:
        Nmax (int): The maximum angular momentum quantum number to include in the sum.

    Returns:
        int: The cumulative number of projection quantum numbers for each angular
            momentum 0 to Nmax.
    """
    # sum([2 * x + 1 for x in range(0, Nmax + 1)]) =
    return int((Nmax + 1) ** 2)


def num_proj(angmom):
    """
    Computes the number of possible projection quantum numbers for a given angular
    momentum quantum number.

    Args:
        angmom (int | HalfInt | float): The angular momentum quantum number for which
            to calculate the number of possible projections.

    Returns:
        int: The number of possible projection quantum numbers for the specified
            angular momentum.
    """
    return int(2 * angmom + 1)


def vector_dot(A, B):
    """Cartesian dot product of two vectors of (matrix) operators A, B

    Args:
        A,B (np.ndarray): Cartesian length-3 vectors of operators (3xNxN)

    Returns:
        np.ndarray: result of the dot product, NxN array.
    """
    return np.einsum("cij,cjk->ik", A, B, optimize="optimal")


def tensor_dot(T1, T2):
    """Product of two rank-2 spherical tensors T1, T2

    A function to calculate the scalar product of two spherical tensors
    T1 and T2 are lists or np arrays that represent the spherical tensors
    lists are indexed from lowest m to highests

    Args:
        T1,T2 (list of np.ndarray) - length-5 list of np.ndarray

    Returns:
        X (np.ndarray) - scalar product of spherical tensors
    """
    x = np.zeros(T1[0].shape, dtype=np.complex128)
    for i, q in enumerate(range(-2, 2 + 1)):
        x += ((-1) ** q) * np.dot(T1[i], T2[-(i + 1)])
    return x


"""
Angular momentum operators
"""


def raising_operator(j):
    """Creates the angular momentum raising operator for j

    In the j,mj basis running from max(mj) to min (mj) creates a matrix that
    represents the operator j+\\|j,mj> = \\|j,mj+1>

    Args:
        j (float) : value of the angular momentum

    Returns:
        J+ (np.ndarray) : Array representing the operator J+,
            has shape ((2j+1),(2j+1))

    """
    matrix_width = num_proj(j)
    J = np.zeros((matrix_width, matrix_width), dtype=float)
    for x, MJ1 in enumerate(proj_iter(j)):
        for y, MJ2 in enumerate(proj_iter(j)):
            if MJ1 == MJ2 + 1:
                J[x, y] = np.sqrt((j - MJ2) * (j + MJ2 + 1))
    return J


# produce the three generalised projections of angular momentum:
# for S=1/2 these should return the Pauli matrices.
# for the source of these definitions see any good QM textbook e.g.
# Bransden & Joachain (or wikipedia)


def x_operator(J):
    """operator for X component of J

    Creates the Cartesian operator Jx for a given J

    Args:
        J (float): Magnitude of angular momentum
    Returns:
        Jx (np.ndarray) : 2J+1 square np array
    """
    J_plus = raising_operator(J)
    J_minus = np.transpose(J_plus)
    return 0.5 * (J_plus + J_minus)


def y_operator(J):
    """operator for Y component of J

    Creates the Cartesian operator Jy for a given J

    Args:
        J (float): Magnitude of angular momentum
    Returns:
        Jy (np.ndarray) : 2J+1 square np array
    """
    J_plus = raising_operator(J)
    J_minus = np.transpose(J_plus)
    return 0.5j * (J_minus - J_plus)


def z_operator(J):
    """operator for Z component of J

    Creates the Cartesian operator Jz for a given J. This is diagonal in the j,mj basis
    such that jz|j,mj> = mj|j,mj>

    Args:
        J (float): Magnitude of angular momentum
    Returns:
        Jz (np.ndarray) : 2J+1 square np array
    """
    J_plus = raising_operator(J)
    J_minus = np.transpose(J_plus)
    return 0.5 * (J_plus @ J_minus - J_minus @ J_plus)


def generate_vecs(Nmax, I1, I2):
    """Build N, I1, I2 angular momentum vectors

    Generate the vectors of the angular momentum operators which we need
    to be able to produce the Hamiltonian

    Args:
        Nmax (float): maximum rotational level to include in calculations
        I1,I2 (float): Nuclear spins of nuclei 1 and 2
    Returns:
        N_vec,I1_vec,I2_vec (list of np.ndarray): length-3 list of
            (2Nmax+1)*(2I1+1)*(2I2+1) square np arrays
    """

    shapeN = num_proj_with_below(Nmax)
    shape1 = num_proj(I1)
    shape2 = num_proj(I2)

    Nx = np.array([[]])
    Ny = np.array([[]])
    Nz = np.array([[]])

    for n in range(0, Nmax + 1):
        Nx = block_diag(Nx, x_operator(n))
        Ny = block_diag(Ny, y_operator(n))
        Nz = block_diag(Nz, z_operator(n))

    # remove the first element of the N vectors, which are empty
    Nx = Nx[1:, :]
    Ny = Ny[1:, :]
    Nz = Nz[1:, :]

    # Each of the following corresponds to the product [N x 1Rb x 1Cs]
    # This gives the operators for N in the full hyperfine space.

    # np.kron is the function for the Kronecker product, often also called
    # the tensor product.

    N_vec = np.array(
        [
            np.kron(Nx, np.kron(np.identity(shape1), np.identity(shape2))),
            np.kron(Ny, np.kron(np.identity(shape1), np.identity(shape2))),
            np.kron(Nz, np.kron(np.identity(shape1), np.identity(shape2))),
        ]
    )

    # we also have to repeat for the nuclear spins
    I1_vec = np.array(
        [
            np.kron(
                np.identity(shapeN),
                np.kron(x_operator(I1), np.identity(shape2)),
            ),
            np.kron(
                np.identity(shapeN),
                np.kron(y_operator(I1), np.identity(shape2)),
            ),
            np.kron(
                np.identity(shapeN),
                np.kron(z_operator(I1), np.identity(shape2)),
            ),
        ]
    )

    I2_vec = np.array(
        [
            np.kron(
                np.identity(shapeN),
                np.kron(np.identity(shape1), x_operator(I2)),
            ),
            np.kron(
                np.identity(shapeN),
                np.kron(np.identity(shape1), y_operator(I2)),
            ),
            np.kron(
                np.identity(shapeN),
                np.kron(np.identity(shape1), z_operator(I2)),
            ),
        ]
    )

    return N_vec, I1_vec, I2_vec


def wigner_D(l, m, alpha, beta, gamma):  # noqa: E741
    """The Wigner D matrix with labels l and m.

    Calculates the Wigner D Matrix for the given Alpha,beta,gamma in radians.
    The wigner-D matrices represent rotations of angular momentum operators.
    The indices l and m determine the value of the matrix.
    The second index (m') is always zero.

    The input angles are the x-z-x euler angles

    Args:
        l (int) : order of wigner Matrix
        m (float): first index of Wigner Matrix
        alpha,beta,gamma (float) : x,z,x Euler angles in radians
    Returns:
        D (float) : Value of the wigner-D matrix
    """
    prefactor = np.sqrt((4 * np.pi) / (2 * l + 1))
    function = np.conj(sph_harm(m, l, alpha, beta))
    return prefactor * function


def T2_C(Nmax, I1, I2):
    """
    The irreducible spherical tensors for the spherical harmonics in the
    rotational basis.

    Args:
        Nmax (int) : Maximum rotational state to include
        I1,I2 (float) :  The nuclear spins of nucleus 1 and 2

    Returns:
        T (list of np.ndarray) : spherical tensor T^2(C). Each element is a
            spherical operator
    """
    matrix_width = num_proj_with_below(Nmax)

    T = np.zeros((5, matrix_width, matrix_width))

    for x, (N, MN) in enumerate(sph_iter(Nmax)):
        for y, (Np, MNp) in enumerate(sph_iter(Nmax)):
            for i, q in enumerate(range(-2, 2 + 1)):
                T[i][x, y] = (
                    ((-1) ** MN)
                    * np.sqrt((2 * N + 1) * (2 * Np + 1))
                    * wigner_3j(N, 2, Np, 0, 0, 0)
                    * wigner_3j(N, 2, Np, float(-MN), q, float(MNp))
                )

    # Expand into full Hyperfine basis
    nuclear_identity_width = num_proj(I1) * num_proj(I2)
    nuclear_identity = np.identity(nuclear_identity_width)
    expanded_matrix_width = matrix_width * nuclear_identity_width
    T_expanded = np.zeros((5, expanded_matrix_width, expanded_matrix_width))
    for i, q in enumerate(range(-2, 2 + 1)):
        T_expanded[i] = np.kron(T[i], nuclear_identity)

    return T_expanded


def makeT2(I1_vec, I2_vec):
    """Construct the spherical tensor T2 from two cartesian vectors of operators.

    Args:
        I1,I2 (list of np.ndarray): Length-3 list of cartesian angular momentum
            operators: the output of makevecs
    Returns:
        T (list of np.ndarray): T^2(I1,I2) length-5 list of spherical angular
            momentum operators
    """
    T2m2 = 0.5 * (
        np.dot(I1_vec[0], I2_vec[0])
        - 1.0j * np.dot(I1_vec[0], I2_vec[1])
        - 1.0j * np.dot(I1_vec[1], I2_vec[0])
        - np.dot(I1_vec[1], I2_vec[1])
    )
    T2p2 = 0.5 * (
        np.dot(I1_vec[0], I2_vec[0])
        + 1.0j * np.dot(I1_vec[0], I2_vec[1])
        + 1.0j * np.dot(I1_vec[1], I2_vec[0])
        - np.dot(I1_vec[1], I2_vec[1])
    )

    T2m1 = 0.5 * (
        np.dot(I1_vec[0], I2_vec[2])
        - 1.0j * np.dot(I1_vec[1], I2_vec[2])
        + np.dot(I1_vec[2], I2_vec[0])
        - 1.0j * np.dot(I1_vec[2], I2_vec[1])
    )
    T2p1 = -0.5 * (
        np.dot(I1_vec[0], I2_vec[2])
        + 1.0j * np.dot(I1_vec[1], I2_vec[2])
        + np.dot(I1_vec[2], I2_vec[0])
        + 1.0j * np.dot(I1_vec[2], I2_vec[1])
    )

    T20 = -np.sqrt(1 / 6) * (
        np.dot(I1_vec[0], I2_vec[0]) + np.dot(I1_vec[1], I2_vec[1])
    ) + np.sqrt(2 / 3) * np.dot(I1_vec[2], I2_vec[2])

    T = [T2m2, T2m1, T20, T2p1, T2p2]

    return T


"""
Molecular operators
"""


def unit_dipole_operator(Nmax, h):
    """
    Generates the induced dipole moment operator for a rigid rotor molecule in
    the spherical harmonic basis.

    This function constructs the dipole moment matrix by iterating over spherical
    harmonics for rotational states up to Nmax. It uses the Wigner 3j symbols to
    calculate the matrix elements, which represent transitions between rotational
    states induced by an external dipole field of specified helicity (h).

    Args:
        Nmax (int): The maximum rotational quantum number to include.
        h (float): The helicity of the dipole field.

    Returns:
        np.ndarray: The induced dipole moment matrix for transitions between
                    rotational states up to Nmax.
    """
    matrix_width = num_proj_with_below(Nmax)
    dmat = np.zeros((matrix_width, matrix_width))

    for i, (N1, M1) in enumerate(sph_iter(Nmax)):
        for j, (N2, M2) in enumerate(sph_iter(Nmax)):
            dmat[i, j] = (
                (-1) ** M1
                * np.sqrt((2 * N1 + 1) * (2 * N2 + 1))
                * wigner_3j(N1, 1, N2, float(-M1), h, float(M2))
                * wigner_3j(N1, 1, N2, 0, 0, 0)
            )

    return dmat


def expanded_unit_dipole_operator(mol, h):
    """
    Expands a dipole operator originally the spherical harmonic basis
    to cover states in the uncoupled hyperfine basis of the molecule.

    The dipole matrix is expanded to the basis including nuclear spin states by taking
    the Kronecker product with the identity matrix of the appropriate size.

    Args:
        mol: A molecular object with Nmax and nuclear spins (Ii) attributes.
        h (float): The helicity of the dipole field.

    Returns:
        np.ndarray: The expanded dipole operator matrix.
    """
    dmat = unit_dipole_operator(mol.Nmax, h)

    nuc_spin_identity = np.identity(num_proj(mol.Ii[0]) * num_proj(mol.Ii[1]))
    dmat_expanded = np.kron(dmat, nuc_spin_identity)

    return dmat_expanded


def electric_gradient(Nmax):
    """
    Calculates the electric field gradient at the nucleus for rotational states.

    Constructs a spherical tensor representing the electric field gradient at the
    nucleus, which depends on the rotational states but not the nuclear spin states.

    Args:
        Nmax (int): The maximum rotational state to include.

    Returns:
        list of np.ndarray: A length-5 list of arrays representing the electric field
                            gradient tensor components.
    """
    matrix_width = num_proj_with_below(Nmax)
    T = np.zeros((5, matrix_width, matrix_width))

    for i, (N1, M1) in enumerate(sph_iter(Nmax)):
        for j, (N2, M2) in enumerate(sph_iter(Nmax)):
            for n, q in enumerate(range(-2, 2 + 1)):
                T[n][i, j] = (
                    (-1) ** M1
                    * np.sqrt((2 * N1 + 1) * (2 * N2 + 1))
                    * wigner_3j(N1, 2, N2, float(-M1), q, float(M2))
                    * wigner_3j(N1, 2, N2, 0, 0, 0)
                )

    return T


def expanded_electric_gradient(mol):
    """Expands the electric gradient tensor to include nuclear spin states.

    Uses the Kronecker product to expand the electric gradient tensor calculated for
    rotational states into a basis that also includes nuclear spin states.

    Args:
        mol: A molecular object with Nmax and nuclear spins (Ii) attributes.

    Returns:
        np.ndarray: The expanded electric gradient tensor.
    """
    elec_mat = electric_gradient(mol.Nmax)

    nuc_spin_basis_size = num_proj(mol.Ii[0]) * num_proj(mol.Ii[1])
    nuc_spin_identity = np.identity(nuc_spin_basis_size)

    expanded_matrix_width = elec_mat[0].shape[0] * nuc_spin_basis_size
    T_expanded = np.empty((5, expanded_matrix_width, expanded_matrix_width))

    for i in range(5):
        T_expanded[i] = np.kron(elec_mat[i], nuc_spin_identity)

    return T_expanded


def quad_moment(I_nuc):
    """Calculates the nuclear electric quadrupole moments of nuclear spin.

    Constructs a spherical tensor representing the nuclear quadrupole moment,
    which depends on the nuclear spin states, not the rotational states.

    Args:
        I_nuc (float): The nuclear spin quantum number.

    Returns:
        list of np.ndarray: A length-5 list of arrays representing the nuclear
            quadrupole moment tensor components.
    """
    matrix_width = num_proj(I_nuc)
    T = np.zeros((5, matrix_width, matrix_width))

    for i, M1 in enumerate(proj_iter(I_nuc)):
        for j, M2 in enumerate(proj_iter(I_nuc)):
            for n, q in enumerate(range(-2, 2 + 1)):
                T[n][i, j] = (
                    (-1) ** int(I_nuc - M1)
                    * wigner_3j(float(I_nuc), 2, float(I_nuc), float(-M1), q, float(M2))
                    / wigner_3j(
                        float(I_nuc), 2, float(I_nuc), float(-I_nuc), 0, float(I_nuc)
                    )
                )

    return T


def expanded_quad_moment(mol, nucleus):
    """Expands the quadrupole moment tensor into the full hyperfine basis.

    Args:
        mol: A molecular object with Nmax and nuclear spins (Ii) attributes.
        nucleus (int): Index of the nucleus (0 or 1) for which to calculate the
                       quadrupole moment tensor.

    Returns:
        np.ndarray: The expanded nuclear quadrupole moment tensor.
    """
    T_nucleus = quad_moment(mol.Ii[nucleus])

    # Expand into full hyperfine basis
    num_N_proj = num_proj_with_below(mol.Nmax)
    nuc_spin_basis_size = num_proj(mol.Ii[0]) * num_proj(mol.Ii[1])
    expanded_matrix_width = num_N_proj * nuc_spin_basis_size

    IdentityN = np.identity(num_N_proj)

    T_exp = np.empty((5, expanded_matrix_width, expanded_matrix_width))

    if nucleus == 0:
        Identity1 = np.identity(num_proj(mol.Ii[1]))
        for i, q in enumerate(range(-2, 2 + 1)):
            T_exp[i] = np.kron(IdentityN, np.kron(T_nucleus[i], Identity1))
    else:
        Identity0 = np.identity(num_proj(mol.Ii[0]))
        for i, q in enumerate(range(-2, 2 + 1)):
            T_exp[i] = np.kron(IdentityN, np.kron(Identity0, T_nucleus[i]))

    return T_exp


def quadrupole(mol):
    """Calculates the nuclear electric quadrupole interaction energy.

    Computes the quadrupole interaction terms for the full hyperfine Hamiltonian using
    spherical tensor algebra. Requires the nuclear quadrupole moments and
    electric field gradients.

    Args:
        mol: A molecular object with necessary attributes like nuclear quadrupole
            moments (Qi), nuclear spins (Ii), and maximum rotational quantum
            number (Nmax).

    Returns:
        np.ndarray: The quadrupole interaction term of the hyperfine Hamiltonian.
    """

    TdE = expanded_electric_gradient(mol)
    Tq0 = expanded_quad_moment(mol, 0)
    Tq1 = expanded_quad_moment(mol, 1)

    if mol.Ii[0] < 1:
        Hq0 = 0
    else:
        Hq0 = mol.Qi[0] * tensor_dot(Tq0, TdE)

    if mol.Ii[1] < 1:
        Hq1 = 0
    else:
        Hq1 = mol.Qi[1] * tensor_dot(Tq1, TdE)

    return (Hq0 + Hq1) / 4


def rotational(N, Brot, Drot):
    """Calculates the hyperfine-free rotational Hamiltonian for a rigid rotor molecule.

    Includes both the rotational constant (Brot) and the centrifugal distortion (Drot).

    Matrix is returned in the N,MN basis with MN going from maximum to minimum.

    Args:
        N (list of np.ndarray): length 3 list representing the Angular momentum
            vector for rotation.
        Brot (float): Rotational constant in joules.
        Drot (float): Centrifugal distortion constant in joules.

    Returns:
        np.ndarray: The rotational Hamiltonian in the N, MN basis.
    """

    N_squared = vector_dot(N, N)
    return Brot * N_squared - Drot * N_squared * N_squared


def zeeman(Cz, J):
    """Calculates the Zeeman effect Hamiltonian for a magnetic field along the z-axis.

    Linear Zeeman shift, fixed magnetic field along z so only need the
    last component of the angular momentum vector.

    Args:
        Cz (float): Zeeman coefficient/magnetic moment.
        J (list of np.ndarray): Angular momentum operator vector.

    Returns:
        np.ndarray: The Zeeman Hamiltonian.
    """
    Hzeeman = -Cz * J[2]
    return Hzeeman


def scalar_nuclear(Ci, J1, J2):
    """Calculate the scalar spin-spin interaction term

    Returns the scalar spin-spin term of the HF Hamiltonian

    Args:
        Ci(float) - Scalar spin coupling coefficient
        J1,J2 (list of np.ndarray) - Angular momentum vectors

    Returns:
        H (np.ndarray) - Hamiltonian for spin-spin interaction
    """
    return Ci * vector_dot(J1, J2)


def tensor_nuclear(C3, I1_vec, I2_vec, I1_val, I2_val, Nmax):
    """Calculate the tensor spin-spin interaction.

    This function is to calculate the tensor spin-spin interaction.
    This version uses spherical tensors to calculate the correct off-diagonal
    behaviour.

    Args:
        C3 (float) - spin-spin coupling constant
        I1,I2 (float) - Cartesian Angular momentum operator Vectors
        Nmax (int) - maximum rotational state to include

    Returns:
        Hss (np.ndarray) - Hamiltonian for tensor spin-spin interaction
    """

    # steps for maths, creates the spherical tensors
    T1 = T2_C(Nmax, I1_val, I2_val)
    T2 = makeT2(I1_vec, I2_vec)
    # return final Hamiltonian
    tensorss = np.sqrt(6) * C3 * tensor_dot(T1, T2)

    return tensorss


def unit_ac_iso(Nmax, I1, I2):
    """Calculate isotropic Stark shifts

    Generates the effect of the isotropic AC Stark shift for a rigid-rotor
    like molecule.

    This term is calculated differently to all of the others in this work
    and is based off Jesus Aldegunde's FORTRAN 77 code. It iterates over
    N,MN,N',MN' to build a matrix without hyperfine structure then uses
    kronecker products to expand it into all of the hyperfine states.

    Args:
        Nmax (int) - maximum rotational quantum number to calculate (int)
        I1,I2 (float) - Nuclear spin of nucleus 1,2

    Returns:
        H (np.ndarray) - isotropic AC Stark Hamiltonian

    """
    matrix_width = num_proj_with_below(Nmax) * num_proj(I1) * num_proj(I2)
    HAC = -1 * np.identity(matrix_width)

    # return the matrix, in the full uncoupled basis.
    return (1.0 / (2 * scipy.constants.epsilon_0 * scipy.constants.c)) * HAC


def unit_ac_aniso(Nmax, I1, I2, Beta):
    """Calculate anisotropic ac stark shift.

    Generates the effect of the anisotropic AC Stark shift for a rigid-rotor
    like molecule.

    This term is calculated differently to all of the others in this work
    and is based off Jesus Aldegunde's FORTRAN 77 code. It iterates over
    N,MN,N',MN' to build a matrix without hyperfine structure then uses
    kronecker products to expand it into all of the hyperfine states.

    Args:

        Nmax (int) - maximum rotational quantum number to calculate
        I1,I2 (float) - Nuclear spin of nucleus 1,2
        Beta (float) - polarisation angle of the laser in Radians

    Returns:
        H (np.ndarray): Hamiltonian in joules
    """

    matrix_width = num_proj_with_below(Nmax)
    HAC = np.zeros((matrix_width, matrix_width), dtype=complex)
    for i, (N1, M1) in enumerate(sph_iter(Nmax)):
        for j, (N2, M2) in enumerate(sph_iter(Nmax)):
            M = M2 - M1
            HAC[i, j] = -1 * (
                wigner_D(2, M, 0, Beta, 0)
                * (-1) ** M2
                * np.sqrt((2 * N1 + 1) * (2 * N2 + 1))
                * wigner_3j(N2, 2, N1, float(-M2), float(M), float(M1))
                * wigner_3j(N2, 2, N1, 0, 0, 0)
            )

    HAC[np.isnan(HAC)] = 0

    HAC_expanded = np.kron(HAC, np.identity(num_proj(I1) * num_proj(I2)))
    # return the matrix, in the full uncoupled basis.
    return (1.0 / (2 * scipy.constants.epsilon_0 * scipy.constants.c)) * HAC_expanded


# Now some functions to take these functions and assemble them into the physical
# Hamiltonians where necessary.


@log_time
def hyperfine_ham(mol):
    """Calculate the field-free Hyperfine hamiltonian

    Wrapper to call all of the functions that are appropriate for the singlet-sigma
    hyperfine hamiltonian.

    Args:
        Nmax (int) - Maximum rotational level to include
        I1_mag,I2_mag (float) - magnitude of the nuclear spins
        Consts (Dictionary): Dict of molecular constants
    Returns:
        H0 : Hamiltonian for the hyperfine structure in joules
    """
    N, I1, I2 = generate_vecs(mol.Nmax, mol.Ii[0], mol.Ii[1])
    rotational_part = rotational(N, mol.Brot, mol.Drot)
    scalar_part = (
        scalar_nuclear(mol.Ci[0], N, I1)
        + scalar_nuclear(mol.Ci[1], N, I2)
        + scalar_nuclear(mol.C4, I1, I2)
    )
    tensor_nuclear_part = tensor_nuclear(mol.C3, I1, I2, mol.Ii[0], mol.Ii[1], mol.Nmax)
    quadrupole_part = quadrupole(mol)
    H = rotational_part + scalar_part + tensor_nuclear_part + quadrupole_part
    return H


@log_time
def dc_ham(mol):
    """calculate HDC for a diatomic molecule

    Generates the effect of the dc Stark shift for a rigid-rotor like
    molecule.

    This term is calculated differently to all of the others in this work
    and is based off Jesus Aldegunde's FORTRAN 77 code. It iterates over
    N,MN,N',MN' to build a matrix without hyperfine structure then uses
    kronecker products to expand it into all of the hyperfine states.


    Args:
        Nmax(int) -  maximum rotational quantum number to calculate
        d0 (float) - Permanent electric dipole momentum
        I1,I2 (float) - Nuclear spin of nucleus 1,2


    Returns:
        H (np.ndarray) - DC Stark Hamiltonian in joules
    """

    return -1 * mol.d0 * expanded_unit_dipole_operator(mol, 0)


@log_time
def zeeman_ham(mol):
    """Assembles the Zeeman term and generates operator vectors

    Calculates the Zeeman effect for a magnetic field on a singlet-sigma molecule.
    There is no electronic term and the magnetic field is fixed to be along the z axis.

    Args:
        Nmax (int) - Maximum rotational level to include
        I1_mag,I2_mag (float) - magnitude of the nuclear spins
        Consts (Dictionary): Dict of molecular constants

    Returns:
        Hz (np.ndarray): Hamiltonian for the zeeman effect
    """
    N, I1, I2 = generate_vecs(mol.Nmax, mol.Ii[0], mol.Ii[1])
    H = zeeman(mol.Mui[0], I1) + zeeman(mol.Mui[1], I2) + zeeman(mol.MuN, N)
    return H


@log_time
def ac_ham(mol, a02, beta=0):
    """
    Computes the AC Stark shift Hamiltonian for a molecule in an oscillating electric
    field.

    The function combines the isotropic and anisotropic AC Stark shifts to obtain the
    total AC Stark Hamiltonian, considering the polarization of the electric field.

    Args:
        mol: A molecule object containing the necessary attributes for calculation.
        a02 (tuple): A two-element tuple containing the isotropic and anisotropic
            polarizabilities.
        beta (float, optional): The polarization angle of the electric field.
            Defaults to 0.

    Returns:
        np.ndarray: The AC Stark shift Hamiltonian matrix.
    """
    Hac = a02[0] * unit_ac_iso(mol.Nmax, mol.Ii[0], mol.Ii[1]) + a02[1] * unit_ac_aniso(
        mol.Nmax, mol.Ii[0], mol.Ii[1], beta
    )
    return Hac



import numpy as np
import matplotlib.pyplot as plt
from matplotlib.collections import LineCollection

from scipy.special import sph_harm
from diatomic.operators import uncoupled_basis_iter


def _make_segments(x, y):
    """
    Create list of line segments from x and y coordinates, in the correct format for
    LineCollection. Has the form:  numlines x (points per line) x 2 (x and y) array
    """

    points = np.array([x, y]).T.reshape(-1, 1, 2)
    segments = np.concatenate([points[:-1], points[1:]], axis=1)

    return segments


def colorline(ax, x, y, colors, linewidth=3):
    """
    Adds a line with colour (and therefore transparency) varying per (x,y) point to ax.
    """
    segments = _make_segments(x, y)
    lc = LineCollection(segments, colors=colors, linewidth=linewidth)
    ax.add_collection(lc)

    return lc


def _surface_plot(ax, fxs, fys, fzs):
    # Add axis lines
    ax_len = 0.5
    ax.plot([-ax_len, ax_len], [0, 0], [0, 0], c="0.5", lw=1, alpha=0.3)
    ax.plot([0, 0], [-ax_len, ax_len], [0, 0], c="0.5", lw=1, alpha=0.3)
    ax.plot([0, 0], [0, 0], [-ax_len, ax_len], c="0.5", lw=1, alpha=0.3)
    # Set axes limits
    ax_lim = 0.5
    ax.set_xlim(-ax_lim, ax_lim)
    ax.set_ylim(-ax_lim, ax_lim)
    ax.set_zlim(-ax_lim, ax_lim)
    # Set camera position
    ax.view_init(elev=15, azim=45)  # Reproduce view
    ax.set_xlim3d(-0.2, 0.2)  # Reproduce magnification
    ax.set_ylim3d(-0.2, 0.2)  # ...
    ax.set_zlim3d(-0.2, 0.2)  # ...
    # Turn off Axes
    ax.axis("off")
    # Draw
    ax.patch.set_alpha(0.0)
    ax.plot_surface(
        fxs,
        fys,
        fzs,
        rstride=1,
        cstride=1,
        cmap=plt.get_cmap("viridis"),
        linewidth=0,
        antialiased=False,
        alpha=0.3,
        shade=False,
    )


def plot_rotational_3d(ax, mol, eigenvector, plot_res=50):
    """
    Plots eigenvectors rotational distribution in 3D, ax's projection type must be '3d'
    """
    if ax.name != "3d":
        raise TypeError("Axis must have projection type of '3d'.")

    # Polar and Azimuthal angles to Sample
    theta = np.linspace(0, np.pi, plot_res)
    phi = np.linspace(0, 2 * np.pi, plot_res)
    # Create a 2-D meshgrid of (theta, phi) angles.
    theta_grid, phi_grid = np.meshgrid(theta, phi)
    # Calculate the unit sphere Cartesian coordinates of each (theta, phi).
    xyz = np.array(
        [
            np.sin(theta_grid) * np.sin(phi_grid),
            np.sin(theta_grid) * np.cos(phi_grid),
            np.cos(theta_grid),
        ]
    )

    f_grid = np.zeros((plot_res, plot_res), dtype=np.cdouble)

    for i, (N, MN, M1, M2) in enumerate(uncoupled_basis_iter(mol.Nmax, *mol.Ii)):
        f_grid += eigenvector[i] * sph_harm(MN, N, phi_grid, theta_grid)

    Yx, Yy, Yz = np.abs(f_grid) ** 2 * xyz  # get final output cartesian coords
    _surface_plot(ax, Yx, Yy, Yz)


def plot_rotational_2d(ax, mol, eigenvector, plot_res=200, format_axes=True):
    """
    Plots eigenvectors rotational distribution in 2d at phi=0,
    ax's projection type must be 'polar'
    """

    # Polar and Azimuthal angles to Sample
    thetas = np.linspace(-np.pi, np.pi, plot_res)
    f_grid = np.zeros((plot_res), dtype=np.cdouble)

    for i, (N, MN, M1, M2) in enumerate(uncoupled_basis_iter(mol.Nmax, *mol.Ii)):
        f_grid += eigenvector[i] * sph_harm(MN, N, 0, np.abs(thetas))

    probs = np.abs(f_grid) ** 2
    xs = np.sin(thetas) * probs
    ys = np.cos(thetas) * probs
    ax.plot(xs, ys)
    ax.axes.set_aspect("equal")
    if format_axes:
        ax.plot([-0.15, 0.15], [0, 0], c="k", alpha=0.2)
        ax.plot([0, 0], [-0.15, 0.15], c="k", alpha=0.2)
        ax.set_axis_off()



import scipy.constants
from diatomic.operators import HalfInt

pi = scipy.constants.pi
h = scipy.constants.h
muN = scipy.constants.physical_constants["nuclear magneton"][0]
bohr = scipy.constants.physical_constants["Bohr radius"][0]
eps0 = scipy.constants.epsilon_0
DebyeSI = 3.33564e-30  # C m


class SingletSigmaMolecule:
    """
    A container to store all the necessary constants of a molecule. This can be passed
    around functions that need to know molecular constants to build hamiltonians, for
    example.
    """

    def __init__(
        self,
        Ii: tuple[int | HalfInt, int | HalfInt] = (0, 0),
        Nmax: int = 1,
        Brot: float = 1e9 * h,
        Drot: float = 0.0,
        Ci: tuple[float, float] = (0.0, 0.0),
        C4: float = 0.0,
        C3: float = 0.0,
        Qi: tuple[float, float] = (0.0, 0.0),
        d0: float = 1 * DebyeSI,
        MuN: float = 0.0,
        Mui: tuple[float, float] = (0.0, 0.0),
        a02: dict[float, tuple[float, float]] = {},
    ):
        # Nuclear Spin magnitudes
        self.Ii = Ii

        # Maximum rotational basis components considered
        self.Nmax = Nmax

        # Rotational Constants
        self.Brot = Brot
        self.Drot = Drot

        # Hyperfine Constants
        self.Ci = Ci  # Scalar Spin-Rotation
        self.C4 = C4  # Scalar Spin-Spin
        self.C3 = C3  # Tensor Spin-Spin
        self.Qi = Qi  # Quadrupole

        # Electric Dipole Moment (DC Stark)
        self.d0 = d0

        # Magnetic Dipole Moments (Zeeman)
        self.MuN = MuN
        self.Mui = Mui

        # Polarisability, function of wavelength (AC Stark)
        self.a02 = a02  # (iso, aniso)

    @classmethod
    def from_preset(cls, str_name):
        """
        Loads the molecule class from a preset of constants defined in the `presets`
        dictionary.
        """
        if str_name in cls.presets:
            return cls(**cls.presets[str_name])
        else:
            raise KeyError("Preset name does not exist.")

    presets = {
        "RigidRotor": {
            "Ii": (0, 0),
            "d0": 1 * DebyeSI,
            "Brot": 1e9 * h,
            "Drot": 0,
            "Qi": (0, 0),
            "Ci": (0, 0),
            "C3": 0,
            "C4": 0,
            "MuN": 0.01 * muN,
            "Mui": (0, 0),
        },
        "Testing": {
            "Ii": (HalfInt(of=3), 1),
            "d0": 1,
            "Brot": 1,
            "Drot": 1,
            "Qi": (1, 1),
            "Ci": (1, 1),
            "C3": 1,
            "C4": 1,
            "MuN": 1,
            "Mui": (1, 1),
            "a02": {1: (1, 1)},
        },
        "Rb87Cs133": {
            # Most recent Rb87Cs133 Constants are given in the supplementary
            # of Gregory et al., Nat. Phys. 17, 1149-1153 (2021)
            # https://www.nature.com/articles/s41567-021-01328-7
            # Polarisabilities are for 1064 nm reported
            # in Blackmore et al., PRA 102, 053316 (2020)
            # https://journals.aps.org/pra/abstract/10.1103/PhysRevA.102.053316
            "Ii": (HalfInt(of=3), HalfInt(of=7)),
            "d0": 1.225 * DebyeSI,
            "Brot": 490.173994326310e6 * h,
            "Drot": 207.3 * h,
            "Qi": (-809.29e3 * h, 59.98e3 * h),
            "Ci": (29.4 * h, 196.8 * h),
            "C3": 192.4 * h,
            "C4": 19.0189557e3 * h,
            "MuN": 0.0062 * muN,
            "Mui": (1.8295 * muN, 0.7331 * muN),
            "a02": {
                1064: (
                    2020 * 4 * pi * eps0 * bohr**3,
                    1997 * 4 * pi * eps0 * bohr**3,
                ),
                1065: (
                    1825 * 4 * pi * eps0 * bohr**3,
                    1981 * 4 * pi * eps0 * bohr**3,
                ),
                817: (
                    443 * 4 * pi * eps0 * bohr**3,
                    -2816 * 4 * pi * eps0 * bohr**3,
                ),
            },
        },
        "K41Cs133": {
            # K41Cs133 values are from theory:
            # Vexiau et al., Int. Rev. Phys. Chem. 36, 709-750 (2017)
            # https://www.tandfonline.com/doi/full/10.1080/0144235X.2017.1351821
            # Aldegunde et al., PRA 96, 042506 (2017)
            # https://journals.aps.org/pra/abstract/10.1103/PhysRevA.96.042506
            "Ii": (HalfInt(of=3), HalfInt(of=7)),
            "d0": 1.84 * DebyeSI,
            "Brot": 880.326e6 * h,
            "Drot": 0 * h,
            "Qi": (-0.221e6 * h, 0.075e6 * h),
            "Ci": (4.5 * h, 370.8 * h),
            "C3": 9.9 * h,
            "C4": 628 * h,
            "MuN": 0.0 * muN,
            "Mui": (0.143 * (1 - 1340.7e-6) * muN, 0.738 * (1 - 6337.1e-6) * muN),
        },
        "K40Rb87": {
            # For K40Rb87:
            # Brot, Q1, Q2 are from Ospelkaus et al., PRL 104, 030402 (2010)
            # https://journals.aps.org/prl/pdf/10.1103/PhysRevLett.104.030402
            # d0 is from Ni et al., Science 322, 231-235 (2008)
            # https://www.science.org/doi/10.1126/science.1163861
            # a0, a2 are from Neyenhuis et al., PRL 109, 230403 (2012)
            # https://journals.aps.org/prl/pdf/10.1103/PhysRevLett.109.230403
            # All other parameters are from Aldegunde et al., PRA 96, 042506 (2017)
            # https://journals.aps.org/pra/abstract/10.1103/PhysRevA.96.042506
            "Ii": (HalfInt(of=4), HalfInt(of=3)),
            "d0": 0.566 * DebyeSI,
            "Brot": 1113.950e6 * h,
            "Drot": 0 * h,
            "Qi": (0.45e6 * h, -1.41e6 * h),
            "Ci": (-24.1 * h, 419.5 * h),
            "C3": -48.2 * h,
            "C4": -2028.8 * h,
            "MuN": 0.0140 * muN,
            "Mui": (-0.324 * (1 - 1321e-6) * muN, 1.834 * (1 - 3469e-6) * muN),
        },
        "Na23K40": {
            # For Na23K40
            # Parameters from Will et al., PRL 116, 225306 (2016)
            # https://journals.aps.org/prl/abstract/10.1103/PhysRevLett.116.225306
            # and from Aldegunde et al., PRA 96, 042506 (2017)
            # https://journals.aps.org/pra/abstract/10.1103/PhysRevA.96.042506
            "Ii": (HalfInt(of=3), HalfInt(of=8)),
            "d0": 2.72 * DebyeSI,
            "Brot": 2.8217297e9 * h,
            "Drot": 0 * h,
            "Qi": (-0.187e6 * h, 0.899e6 * h),
            "Ci": (117.4 * h, -97.0 * h),
            "C3": -48.4 * h,
            "C4": -409 * h,
            "MuN": 0.0253 * muN,
            "Mui": (1.477 * (1 - 624.4e-6) * muN, -0.324 * (1 - 1297.4e-6) * muN),
        },
        "Na23Rb87": {
            # Parameters from Guo et al., PRA 97, 020501(R) (2018)
            # https://journals.aps.org/pra/abstract/10.1103/PhysRevA.97.020501
            # and from Aldegunde et al., PRA 96, 042506 (2017)
            # https://journals.aps.org/pra/abstract/10.1103/PhysRevA.96.042506
            "Ii": (HalfInt(of=3), HalfInt(of=3)),
            "d0": 3.2 * DebyeSI,
            "Brot": 2.0896628e9 * h,
            "Drot": 0 * h,
            "Qi": (-0.139e6 * h, -3.048e6 * h),
            "Ci": (60.7 * h, 983.8 * h),
            "C3": 259.3 * h,
            "C4": 6.56e3 * h,
            "MuN": 0.001 * muN,
            "Mui": (1.484 * muN, 1.832 * muN),
        },
        "Na23Cs133": {
            # from Aldegunde et al., PRA 96, 042506 (2017)
            # https://journals.aps.org/pra/abstract/10.1103/PhysRevA.96.042506
            "Ii": (HalfInt(of=3), HalfInt(of=7)),
            "d0": 4.53 * DebyeSI,  # Theory value
            "Brot": 0.962e9 * h,
            "Drot": 0 * h,
            "Qi": (-0.097e6 * h, 0.150e6 * h),
            "Ci": (14.2 * h, 854.5 * h),
            "C3": 105.6 * h,
            "C4": 3941.8 * h,
            "MuN": 0.0 * muN,  # No value
            "Mui": (1.478 * (1 - 639.2e-6) * muN, 0.738 * (1 - 6278.7e-6) * muN),
        },
    }



import numpy as np
from diatomic.systems import SingletSigmaMolecule
import diatomic.operators as operators

from diatomic.calculate import (
    _get_prior_repeat_indices,
    label_states,
    transition_electric_moments,
    _matrix_prod_diagonal,
    _solve_quadratic,
)


def test_solve_quadratic():
    a, b, c = 1, -3, 2  # = (x-1)(x-2)
    assert _solve_quadratic(a, b, c) == 2.0

    # = (x-[1,2,3])(x-2)
    a, b, c = 1, np.array([-3, -4, -5]), np.array([2, 4, 6])
    assert (_solve_quadratic(a, b, c) == np.array([2.0, 2.0, 3.0])).all()


def test_get_prior_repeat_indices():
    test_labels = np.array(
        [[1, 1], [2, 2], [3, 3], [3, 3], [3, 2], [2, 2], [5, 5], [3, 3], [4, 4]]
    )
    expected_output = np.array([0, 0, 0, 1, 0, 1, 0, 2, 0])

    assert (_get_prior_repeat_indices(test_labels) == expected_output).all()


def test_label_states():
    mol = SingletSigmaMolecule.from_preset("Testing")
    mol.Nmax = 2
    mol.Ii = (operators.HalfInt(of=3), 1)

    basis_size = (
        operators.num_proj_with_below(mol.Nmax)
        * operators.num_proj(mol.Ii[0])
        * operators.num_proj(mol.Ii[1])
    )

    states = np.zeros((basis_size, 3))

    N_state = np.zeros((operators.num_proj_with_below(mol.Nmax)))
    N_state[0] = 1  # N=0,MN=0

    i1_state = np.zeros((operators.num_proj(mol.Ii[0])))
    i1_state[1] = 1  # MI1=1/2

    i2_state = np.zeros((operators.num_proj(mol.Ii[1])))
    i2_state[1] = 1  # MI2 = 0

    state = np.kron(N_state, np.kron(i1_state, i2_state))

    states[:, 0] = state

    ##############

    N_state = np.zeros((operators.num_proj_with_below(mol.Nmax)))
    N_state[1] = 1  # N=1,MN=1

    i1_state = np.zeros((operators.num_proj(mol.Ii[0])))
    i1_state[0] = 1  # MI1=3/2

    i2_state = np.zeros((operators.num_proj(mol.Ii[1])))
    i2_state[2] = 1  # MI2 = -1

    state = np.kron(N_state, np.kron(i1_state, i2_state))

    states[:, 1] = state

    ################

    N_state = np.zeros((operators.num_proj_with_below(mol.Nmax)))
    N_state[0] = 1  # N=0,MN=0

    i1_state = np.zeros((operators.num_proj(mol.Ii[0])))
    i1_state[1] = 1  # MI1=1/2

    i2_state = np.zeros((operators.num_proj(mol.Ii[1])))
    i2_state[1] = 1  # MI2 = 0

    state = np.kron(N_state, np.kron(i1_state, i2_state))

    states[:, 2] = state

    labels = label_states(mol, states, ["N", "MN", "MI1", "MI2"])

    expected = np.array(
        [
            [0, 0, operators.HalfInt(of=1), 0],
            [1, 1, operators.HalfInt(of=3), -1],
            [0, 0, operators.HalfInt(of=1), 0],
        ]
    )

    assert (labels == expected).all()

    expected_count = np.array(
        [
            [0, 0, operators.HalfInt(of=1), 0, 0],
            [1, 1, operators.HalfInt(of=3), -1, 0],
            [0, 0, operators.HalfInt(of=1), 0, 1],
        ]
    )

    labels_count = label_states(
        mol, states, ["N", "MN", "MI1", "MI2"], index_repeats=True
    )
    assert (labels_count == expected_count).all()


def test_transition_electric_moments():
    mol = SingletSigmaMolecule.from_preset("Testing")
    mol.Nmax = 1
    mol.Ii = (0, 0)

    d0_matrix = transition_electric_moments(mol, np.identity(4), h=0)
    expected = np.sqrt(1 / 3) * np.array(
        [
            [0, 0, 1, 0],
            [0, 0, 0, 0],
            [1, 0, 0, 0],
            [0, 0, 0, 0],
        ]
    )
    assert np.allclose(d0_matrix, expected)

    d0_from_0 = transition_electric_moments(mol, np.identity(4), h=0, from_states=[0])
    expected = np.array([0, 0, np.sqrt(1 / 3), 0])
    assert np.allclose(d0_from_0, expected)

    d0_from_0 = transition_electric_moments(mol, np.identity(4), h=0, from_states=0)
    expected = np.array([0, 0, np.sqrt(1 / 3), 0])
    assert np.allclose(d0_from_0, expected)

    d0_from_1 = transition_electric_moments(mol, np.identity(4), h=0, from_states=1)
    expected = np.array([0, 0, 0, 0])
    assert np.allclose(d0_from_1, expected)

    d0_from_01 = transition_electric_moments(
        mol, np.identity(4), h=0, from_states=[0, 2]
    )
    expected = np.sqrt(1 / 3) * np.array([[0, 0, 1, 0], [1, 0, 0, 0]])
    assert np.allclose(d0_from_01, expected)

    d0_from_011 = transition_electric_moments(mol, np.identity(4), h=1)
    expected = np.sqrt(1 / 3) * np.array(
        [
            [0, 1, 0, 0],
            [1, 0, 0, 0],
            [0, 0, 0, 0],
            [0, 0, 0, 0],
        ]
    )
    assert np.allclose(d0_from_011, expected)

    d0_from_01m1 = transition_electric_moments(mol, np.identity(4), h=-1, from_states=0)
    expected = np.sqrt(1 / 3) * np.array([0, 0, 0, 1])
    assert np.allclose(d0_from_01m1, expected)

    d0_from_011 = transition_electric_moments(mol, np.identity(4), h=1, from_states=0)
    expected = np.sqrt(1 / 3) * np.array([0, 1, 0, 0])
    assert np.allclose(d0_from_011, expected)

    d0_from_011 = transition_electric_moments(
        mol, np.identity(4), h=1, from_states=[0, 1]
    )
    expected = np.sqrt(1 / 3) * np.array([[0, 1, 0, 0], [1, 0, 0, 0]])
    assert np.allclose(d0_from_011, expected)

    d0_from_011 = transition_electric_moments(
        mol, np.eye(4, 2), h=1, from_states=[0, 1]
    )
    expected = np.sqrt(1 / 3) * np.array([[0, 1], [1, 0]])
    assert np.allclose(d0_from_011, expected)

    eigvecs = np.array([[1.0, 0.0], [0.0, 0.0], [0.0, 0.0], [0.0, 1.0]])
    d0_from_011 = transition_electric_moments(mol, eigvecs, h=-1, from_states=[0, 1])
    expected = np.sqrt(1 / 3) * np.array([[0, 1], [1, 0]])
    assert np.allclose(d0_from_011, expected)

    d0_from_011 = transition_electric_moments(mol, eigvecs, h=1, from_states=[0, 1])
    expected = np.sqrt(1 / 3) * np.array([[0, 0], [0, 0]])
    assert np.allclose(d0_from_011, expected)


def test_diagonal_2d():
    np.random.seed(0)  # Setting a seed for reproducibility
    A = np.random.rand(5, 5)
    B = np.random.rand(5, 5)

    for d in range(-2, 2):
        expected_diag = np.diag(A @ B, k=d)
        result_diag = _matrix_prod_diagonal(A, B, d)
        np.testing.assert_array_almost_equal(result_diag, expected_diag)


def test_diagonal_multidimensional():
    np.random.seed(0)
    A = np.random.rand(3, 6, 6)
    B = np.random.rand(3, 6, 6)

    for d in range(-2, 2):
        expected_diag = np.array([np.diag(a @ b, k=d) for a, b in zip(A, B)])
        result_diag = _matrix_prod_diagonal(A, B, d)
        np.testing.assert_array_almost_equal(result_diag, expected_diag)


def test_diagonal_multidimensional_not_square():
    np.random.seed(0)
    A = np.random.rand(3, 6, 4)
    B = np.random.rand(3, 4, 6)

    for d in range(-2, 2):
        expected_diag = np.array([np.diag(a @ b, k=d) for a, b in zip(A, B)])
        result_diag = _matrix_prod_diagonal(A, B, d)
        np.testing.assert_array_almost_equal(result_diag, expected_diag)



import pytest

from diatomic.operators import (
    HalfInt,
    proj_iter,
    num_proj_with_below,
    num_proj,
    sph_iter,
    uncoupled_basis_pos,
)


def test_constructor():
    hi = HalfInt(of=3)
    assert hi._double == 3

    hi2 = HalfInt(of=4)
    assert hi2 == 2

    with pytest.raises(TypeError):
        HalfInt(of="not an integer")

    with pytest.raises(TypeError):
        HalfInt(of=3.0)


def test_str():
    hi = HalfInt(of=3)
    assert str(hi) == "(3/2)"


def test_repr():
    hi = HalfInt(of=3)
    assert repr(hi) == "((3/2) : HalfInt)"


def test_eq():
    hi1 = HalfInt(of=3)
    hi2 = HalfInt(of=3)

    assert hi1 == hi2
    assert hi1 != HalfInt(of=2)
    assert hi1 != HalfInt(of=5)
    assert hi1 == 1.5
    assert hi1 != 2
    assert hi1 != 2.0


def test_ordering():
    hi1 = HalfInt(of=3)
    hi2 = HalfInt(of=4)

    assert hi1 < hi2
    assert hi1 <= hi2
    assert hi2 > hi1
    assert hi2 >= hi1

    assert hi1 < 2
    assert hi1 <= 2
    assert hi2 > 1
    assert hi2 >= 1


def test_numeric_conversion():
    assert float(HalfInt(of=-3)) == -1.5
    assert float(HalfInt(of=5)) == 2.5
    assert float(HalfInt(of=6)) == 3.0
    assert int(HalfInt(of=-3)) == int(-3 / 2)
    assert int(HalfInt(of=3)) == int(3 / 2)
    assert int(HalfInt(of=4)) == 2
    assert abs(HalfInt(of=3)) == HalfInt(of=3)
    assert abs(HalfInt(of=-3)) == HalfInt(of=3)


def test_add():
    assert HalfInt(of=3) + HalfInt(of=4) == HalfInt(of=7)
    assert HalfInt(of=3) + HalfInt(of=5) == 4
    assert HalfInt(of=3) + 1 == HalfInt(of=5)
    assert HalfInt(of=3) + 1.5 == 3.0


def test_sub():
    assert HalfInt(of=3) - HalfInt(of=4) == HalfInt(of=-1)
    assert HalfInt(of=3) - HalfInt(of=3) == 0
    assert HalfInt(of=3) - 1 == HalfInt(of=1)
    assert HalfInt(of=3) - 0.5 == 1.0


def test_neg():
    assert -HalfInt(of=3) == HalfInt(of=-3)


def test_mul():
    assert HalfInt(of=3) * HalfInt(of=4) == 3.0
    assert HalfInt(of=3) * 2 == 3
    assert HalfInt(of=3) * 3 == HalfInt(of=9)
    assert HalfInt(of=3) * 4 == 6
    assert HalfInt(of=3) * 1.5 == 2.25


def test_num_proj():
    assert num_proj(3) == 7
    assert num_proj(3.0) == 7
    assert num_proj(HalfInt(of=6)) == 7

    assert num_proj(3.5) == 8
    assert num_proj(HalfInt(of=7)) == 8


def test_num_proj_with_below():
    assert num_proj_with_below(0) == 1
    assert num_proj_with_below(1) == 4
    assert num_proj_with_below(2) == 9

    assert num_proj_with_below(0.0) == 1
    assert num_proj_with_below(1.0) == 4
    assert num_proj_with_below(2.0) == 9


def test_proj_iter():
    assert list(proj_iter(0)) == [0]
    assert list(proj_iter(2)) == [2, 1, 0, -1, -2]

    assert list(proj_iter(0.0)) == [0]
    assert list(proj_iter(2.0)) == [2, 1, 0, -1, -2]

    assert list(proj_iter(HalfInt(of=3))) == [
        HalfInt(of=3),
        HalfInt(of=1),
        HalfInt(of=-1),
        HalfInt(of=-3),
    ]


def test_sph_iter():
    assert list(sph_iter(1)) == [(0, 0), (1, 1), (1, 0), (1, -1)]


def test_uncoupled_basis_pos():
    I1, I2 = 3, HalfInt(of=7)
    N, MN, MI1, MI2 = 1, 0, 1, HalfInt(of=-1)
    pos = uncoupled_basis_pos(N, MN, MI1, MI2, I1, I2)
    manual_pos = 0
    for lN, lMN in sph_iter(3):
        for lMI1 in proj_iter(I1):
            for lMI2 in proj_iter(I2):
                if (N, MN, MI1, MI2) == (lN, lMN, lMI1, lMI2):
                    assert manual_pos == pos
                    return
                manual_pos += 1



from diatomic import log_time
import diatomic.operators as operators

import numpy as np
import warnings


@log_time
def solve_system(hamiltonians, num_diagonals=None):
    """
    Solves for the eigenenergies and eigenstates of given Hamiltonian(s over 1
    parameter).

    This function diagonalizes the input Hamiltonian(s) to find the corresponding
    eigenenergies and eigenstates. It supports both single Hamiltonian and a set
    of Hamiltonians (e.g., varying with time or another parameter).

    It takes care to make sure the eigenenergies and eigenstates vary smoothly with
    respect to the varying parameter, unlike `np.eigh`.

    Args:
        hamiltonians (np.ndarray): A single Hamiltonian matrix or an array of
            Hamiltonian matrices.
        num_diagonals (int | None, optional): The number of local swaps above and below
            to consider when smoothing eigenfunctions. Changing this to a number can
            speed up smoothing calculations for large numbers of eigenstates. If None,
            compare all eigenstates to all other eigenstates when smoothing (safest).

    Returns:
        tuple: (eigenenergies, eigenstates).

    Raises:
        ValueError: If the input Hamiltonian has more than three dimensions.
    """
    eigenenergies_raw, eigenstates_raw = log_time(np.linalg.eigh)(hamiltonians)
    if hamiltonians.ndim == 2:
        return eigenenergies_raw, eigenstates_raw
    elif hamiltonians.ndim == 3:
        eigenenergies, eigenstates = sort_smooth(
            eigenenergies_raw, eigenstates_raw, num_diagonals=num_diagonals
        )
        return eigenenergies, eigenstates
    else:
        raise ValueError(
            "Too many dimensions, solve_system doesn't support smoothing"
            " eigenvalues over >1 parameters"
        )


def _matrix_prod_diagonal(A, B, d=0):
    """
    Computes the dth diagonal of the product of two matrices A and B,
    without computing the entire matrix product.

    Args:
        A (np.ndarray): The first matrix operand.
        B (np.ndarray): The second matrix operand.
        d (int, optional): The diagonal offset. Defaults to 0, which gives the
            main diagonal.

    Returns:
        np.ndarray: The computed diagonal elements as a one-dimensional array.
    """
    A_slice = slice(-d if d < 0 else None, -d if d > 0 else None)
    B_slice = slice(d if d > 0 else None, d if d < 0 else None)

    diag = np.einsum("...ab,...ba->...a", A[..., A_slice, :], B[..., B_slice])

    return diag


@log_time
def sort_smooth(eigvals, eigvecs, num_diagonals=None):
    """
    Smooths the eigenvalue trajectories as a function of an external parameter, rather
    than by energy order, as is the case with `np.eigh`.

    This function sorts the eigenvalues and eigenvectors to maintain continuity as the
    external parameter changes, preventing sudden jumps in the eigenvalue ordering.

    Args:
        eigvals (np.ndarray): Eigenvalues array with shape (steps, eigstate).
        eigvecs (np.ndarray): Eigenvectors array with shape (steps, basis, eigstate).
        num_diagonals (int | None, optional): The number of local swaps above and below
            to consider when smoothing eigenfunctions. Changing this to a number can
            speed up smoothing calculations for large numbers of eigenstates. If None,
            compare all eigenstates to all other eigenstates when smoothing (safest).

    Returns:
        tuple: The smoothed eigenvalues and eigenvectors.

    Raises:
        ValueError: If eigvecs does not have three dimensions.
    """
    if eigvecs.ndim != 3:
        raise ValueError("eigvecs is not a set of eigenvectors over one parameter.")

    param_step_count = eigvecs.shape[0]
    basis_size = eigvecs.shape[1]
    eigstate_count = eigvecs.shape[2]

    # Compute overlap matrix between (k)th and (k-1)th
    if num_diagonals is not None:
        k = min(num_diagonals, eigstate_count - 1)
        diagonals = np.zeros((param_step_count - 1, 2 * k + 1, eigstate_count))
        for diag_num in range(-k, k + 1):
            my_prod_diag = np.abs(
                _matrix_prod_diagonal(
                    eigvecs[:-1].swapaxes(-1, -2), eigvecs[1:].conj(), d=diag_num
                )
            )
            if diag_num > 0:
                diagonals[
                    :, k - diag_num, diag_num : diag_num + my_prod_diag.shape[1]
                ] = my_prod_diag
            else:
                diagonals[:, k - diag_num, 0 : my_prod_diag.shape[1]] = my_prod_diag

        best_overlaps = np.argmax(diagonals, axis=-2) - np.arange(
            k, k - eigstate_count, -1
        )
    else:
        overlap_matrices = np.abs(eigvecs[:-1].swapaxes(-1, -2) @ eigvecs[1:].conj())
        best_overlaps = np.argmax(overlap_matrices, axis=1)

    # Cumulative permutations
    integrated_permutations = np.empty((param_step_count, eigstate_count), dtype=int)
    integrated_permutations[-1] = np.arange(eigstate_count)

    for i in range(param_step_count - 2, -1, -1):
        integrated_permutations[i] = best_overlaps[i][integrated_permutations[i + 1]]

    # Rearrange to maintain continuity
    sorted_eigvals = eigvals[
        np.arange(param_step_count)[:, None], integrated_permutations
    ]
    sorted_eigvecs = eigvecs[
        np.arange(param_step_count)[:, None, None],
        np.arange(basis_size)[None, :, None],
        integrated_permutations[:, None, :],
    ]

    return sorted_eigvals, sorted_eigvecs


def _get_prior_repeat_indices(labels):
    """
    Given a numpy array, returns a list of 'prior repeat indices'. Each entry
    of the array that repeats gets an incrementing index on each subsequent occurrence.

    For example, given labels = [[1,1],[2,3],[1,1]], we see [1,1] repeats, and its
    second occurrence gets a 'prior repeat index' of 1. The output is [0,0,1].

    Args:
        labels (np.array): 2D array with labels.

    Returns:
        np.array: 1D array with 'prior repeat indices'.
    """

    double_labels = (2 * labels).astype(int)

    # Identify unique labels, get inverse indices and counts of labels
    _, inverse_indices, counts = np.unique(
        double_labels, return_inverse=True, return_counts=True, axis=0
    )

    # Initialize array for repeat indices
    repeat_indices = np.zeros(double_labels.shape[0], dtype=int)

    # For each unique label
    for idx, count in enumerate(counts):
        # Find all instances of the unique label
        instances = np.nonzero(inverse_indices == idx)[0]

        # Assign repeat indices in order of occurrence
        repeat_indices[instances] = np.arange(count)

    return repeat_indices


def _solve_quadratic(a, b, c):
    """Solve a quadratic equation a*x^2+b*x+c=0 .

    This is a simple function to solve the quadratic formula for x.
    Returns the most positive value of x supported.

    Args:
        a,b,c (floats) - coefficients in quadratic

    Returns:
        x (float) - maximum value of x supported by equation

    """
    x1 = (-b + np.sqrt((b**2) - (4 * (a * c)))) / (2 * a)
    x2 = (-b - np.sqrt((b**2) - (4 * (a * c)))) / (2 * a)
    return np.maximum(x1, x2)


@log_time
def label_states(mol, eigstates, labels, index_repeats=False):
    """
    Labels the eigenstates of a molecule with quantum numbers corresponding to
    specified labels. The returned numbers will only be good if the state is
    well-represented in the desired basis.

    This function processes a list of labels that specify which quantum numbers to
    calculate for the given eigenstates of a molecule. It uses the angular momentum
    operators for the molecule and the eigenstates to calculate these quantum numbers.

    The function supports labels for the total angular momentum vectors
    (N, F, I, I1, I2) for the z-component of them respectively (M<>).
    It also has an option to return the indices of repeated labels.

    Args:
        mol: An object representing the molecule, which contains the maximum angular
            momentum (Nmax) and intrinsic angular momenta (Ii).
        eigstates (ndarray): An array of eigenstates for which the labels are to be
            calculated.
        labels (list of str): A list of strings indicating which quantum numbers to
            calculate. Valid labels are "MN", "MF", "MI", "MI1", "MI2" for M-components
            and "N", "F", "I", "I1", "I2" for squared total angular momentum vectors.
        index_repeats (bool, optional): If True, includes the label indices of repeated
            labels in the output. Defaults to False.

    Returns:
        ndarray: A two-dimensional array where each row corresponds to the desired
            quantum numbers for each eigenstate, and each column corresponds to a label.
    """

    N_op, I1_op, I2_op = operators.generate_vecs(mol.Nmax, mol.Ii[0], mol.Ii[1])

    out_labels = []

    for label in labels:
        if label[0] == "M":
            match label:
                case "MN":
                    Tz = N_op[2]
                case "MF":
                    F = N_op + I1_op + I2_op
                    Tz = F[2]
                case "MI":
                    Itot = I1_op + I2_op
                    Tz = Itot[2]
                case "MI1":
                    Tz = I1_op[2]
                case "MI2":
                    Tz = I2_op[2]
                case _:
                    warnings.warn("Invalid label, ignoring")
                    continue
            double_M_labels = np.rint(
                2 * np.einsum("ik,ij,jk->k", np.conj(eigstates), Tz, eigstates)
            ).real.astype(int)
            M_labels = np.array([operators.HalfInt(of=dl) for dl in double_M_labels])
            out_labels.append(M_labels)
        else:
            match label:
                case "N":
                    T2 = operators.vector_dot(N_op, N_op)
                case "F":
                    F = N_op + I1_op + I2_op
                    T2 = operators.vector_dot(F, F)
                case "I":
                    Itot = I1_op + I2_op
                    T2 = operators.vector_dot(Itot, Itot)
                case "I1":
                    T2 = operators.vector_dot(I1_op, I1_op)
                case "I2":
                    T2 = operators.vector_dot(I2_op, I2_op)
                case _:
                    warnings.warn("Invalid label, ignoring")
                    continue
            T2eigval = np.einsum("ik,ij,jk->k", np.conj(eigstates), T2, eigstates)
            double_T_labels = np.rint(
                2 * _solve_quadratic(1, 1, -1 * T2eigval)
            ).real.astype(int)
            T_labels = np.array([operators.HalfInt(of=dl) for dl in double_T_labels])
            out_labels.append(T_labels)

    if index_repeats:
        out_labels.append(_get_prior_repeat_indices(np.array(out_labels).T))

    return np.array(out_labels).T


def sort_by_labels(eigenlabels, eigenenergies, eigenstates):
    """
    Sorts eigenstates and eigenenergies according to the provided labels
    lexicographically.

    This function sorts the eigenstates and eigenenergies in ascending order based
    on the eigenlabels, which can be quantum numbers or any other identifiers for
    the states. They are sorted with the first label value taking preference, and then
    the second if the first drew, as with a language dictionary.

    Args:
        eigenlabels (np.ndarray): Array of labels corresponding to eigenstates.
        eigenenergies (np.ndarray): Array of eigenenergies to be sorted.
        eigenstates (np.ndarray): Array of eigenstates to be sorted.

    Returns:
        tuple: The sorted eigenlabels, eigenenergies, and eigenstates.
    """
    indices = np.lexsort(eigenlabels[:, ::-1].T)
    return eigenlabels[indices], eigenenergies[:, indices], eigenstates[:, :, indices]


@log_time
def transition_electric_moments(mol, eigenstates, h=0, from_states=slice(None)):
    """
    Calculates the electric dipole transition coupling strengths between molecular
    eigenstates.

    Args:
        eigenstates (ndarray): An array containing the eigenstates of the molecule.
        mol: An object representing the molecule.
        h (int, optional): The helicity of the transition, +1 for sigma+, 0 for pi,
            -1 for sigma-.
        from_states (int, slice, or list, optional): The indices of the initial states
            from which transitions are considered. If an integer is passed, it is
            internallyconverted to a list. By default, all states are
            considered (slice(None)).

    Returns:
        ndarray: An array containing the absolute values of the transition electric
            moments for the specified transitions between eigenstates.
    """

    if isinstance(from_states, int):
        from_states = [from_states]

    dipole_op = mol.d0 * operators.expanded_unit_dipole_operator(mol, h)

    dipole_matrix_elements = (
        eigenstates[..., from_states].conj().swapaxes(-1, -2) @ dipole_op @ eigenstates
    )

    if h != 0:
        """
        After rotating wave approximation for sigma +/-, interaction matrix elements
        have d_pm matrix elements appearing above and below the diagonal respectively.
        """
        dipole_op_other = -dipole_op.conj().T

        opposite_dipole_matrix_elements = (
            eigenstates[..., from_states].conj().swapaxes(-1, -2)
            @ dipole_op_other
            @ eigenstates
        )

        all_mask = np.ones(dipole_matrix_elements.shape)
        mask = np.tril(all_mask, k=-1)
        mask_other = np.triu(all_mask, k=1)

        transition_electric_moments = (
            dipole_matrix_elements * mask + opposite_dipole_matrix_elements * mask_other
        )
    else:
        transition_electric_moments = dipole_matrix_elements

    return np.abs(transition_electric_moments)


def magnetic_moment(mol, eigstates):
    """
    Calculates the magnetic moments of each eigenstate.

    Args:
        eigstates (np.ndarray): array of eigenstates from diagonalisation
        mol: The molecule used to generate the eigenstates

    Returns:
        ndarray: An array corresponding to the magnetic moments of the eigenstates.
    """

    muz = -1 * operators.zeeman_ham(mol)

    mu = np.einsum(
        "vae,ab,vbe->ve", np.conjugate(eigstates), muz, eigstates, optimize="optimal"
    )
    return mu.real


def electric_moment(mol, eigstates):
    """
    Calculates the electric moments of each eigenstate.

    Args:
        eigstates (np.ndarray): array of eigenstates from diagonalisation
        mol: The molecule used to generate the eigenstates

    Returns:
        ndarray: An array corresponding to the electric moments of the eigenstates.
    """

    dz = -1 * operators.dc_ham(mol)

    d = np.einsum(
        "vae,ab,vbe->ve", np.conjugate(eigstates), dz, eigstates, optimize="optimal"
    )
    return d.real



from functools import wraps
import logging
import time

logger = logging.getLogger(__name__)


def configure_logging(level=logging.INFO, filename=None):
    logger = logging.getLogger(__name__)
    logger.handlers = []  # remove existing handlers
    if filename:
        handler = logging.FileHandler(filename)
    else:
        handler = logging.StreamHandler()
    formatter = logging.Formatter(
        "%(asctime)s [%(levelname)s] - %(message)s", datefmt="%Y-%m-%d %H:%M:%S"
    )
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.setLevel(level)
    # Propagate this logger's level to all descendant loggers in the hierarchy.
    logger.propagate = False


def log_time(func):
    """
    Decorator that times the execution of a function while executing and outputs to the
    logger.
    """

    @wraps(func)
    def log_time_wrapper(*args, **kwargs):
        logger.info(f"Starting Function {func.__name__}...")
        start_time = time.perf_counter()
        result = func(*args, **kwargs)
        end_time = time.perf_counter()
        total_time = end_time - start_time
        logger.info(f"Function {func.__name__} took {total_time:.4f} seconds")
        return result

    return log_time_wrapper
